import concurrent
import multiprocessing
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, Future
import concurrent.futures
from contextlib import suppress
from logging import getLogger
from queue import Queue
from threading import Event
from time import sleep
from typing import Dict, Optional, List, Any, Tuple, Set
from pathlib import Path

import pkg_resources
import yaml
from resotoclient import Kind, Model
from resotolib.args import Namespace
from resotolib.baseplugin import BaseCollectorPlugin
from resotolib.baseresources import BaseResource, EdgeType
from resotolib.config import Config
from resotolib.core.actions import CoreFeedback
from resotolib.core.model_export import node_to_dict
from resotolib.json import from_json
from resotolib.proc import emergency_shutdown
from resotolib.types import Json
from rich import print as rich_print
from rich.live import Live
from sqlalchemy.engine import Engine


from cloud2sql.analytics import AnalyticsEventSender
from cloud2sql.show_progress import CollectInfo
from cloud2sql.sql import SqlUpdater, sql_updater

try:
    from cloud2sql.parquet import ParquetModel, ParquetWriter
except ImportError:
    pass


log = getLogger("resoto.cloud2sql")


def collectors(raw_config: Json, feedback: CoreFeedback) -> Dict[str, BaseCollectorPlugin]:
    result = {}
    config: Config = Config  # type: ignore
    for entry_point in pkg_resources.iter_entry_points("resoto.plugins"):
        plugin_class = entry_point.load()
        if issubclass(plugin_class, BaseCollectorPlugin) and plugin_class.cloud in raw_config:
            log.info(f"Found collector {plugin_class.cloud} ({plugin_class.__name__})")
            plugin_class.add_config(config)
            plugin = plugin_class()
            if hasattr(plugin, "core_feedback"):
                setattr(plugin, "core_feedback", feedback.with_context(plugin.cloud))
            result[plugin_class.cloud] = plugin

    Config.init_default_config()
    Config.running_config.data = {**Config.running_config.data, **Config.read_config(raw_config)}
    return result


def configure(path_to_config: Optional[str]) -> Json:
    config = {}
    if path_to_config:
        with open(path_to_config) as f:
            config = yaml.safe_load(f)  # type: ignore

    if "sources" not in config:
        raise ValueError("No sources configured")
    if "destinations" not in config:
        raise ValueError("No destinations configured")

    return config


def collect(
    collector: BaseCollectorPlugin, engine: Optional[Engine], feedback: CoreFeedback, args: Namespace, config: Json
) -> Tuple[str, int, int]:
    if engine:
        return collect_sql(collector, engine, feedback, args)
    else:
        return collect_parquet(collector, feedback, config)


def prepare_node(node: BaseResource, collector: BaseCollectorPlugin) -> Json:
    node._graph = collector.graph
    exported = node_to_dict(node)
    exported["type"] = "node"
    exported["ancestors"] = {
        "cloud": {"reported": {"id": node.cloud().name}},
        "account": {"reported": {"id": node.account().name}},
        "region": {"reported": {"id": node.region().name}},
        "zone": {"reported": {"id": node.zone().name}},
    }
    return exported


def collect_parquet(collector: BaseCollectorPlugin, feedback: CoreFeedback, config: Json) -> Tuple[str, int, int]:
    # collect cloud data
    feedback.progress_done(collector.cloud, 0, 1)
    collector.collect()
    # read the kinds created from this collector
    kinds = [from_json(m, Kind) for m in collector.graph.export_model(walk_subclasses=False)]
    model = ParquetModel(Model({k.fqn: k for k in kinds}))
    node_edge_count = len(collector.graph.nodes) + len(collector.graph.edges)
    ne_current = 0
    progress_update = node_edge_count // 100
    feedback.progress_done("sync_db", 0, node_edge_count, context=[collector.cloud])

    # group all edges by kind of from/to
    edges_by_kind: Set[Tuple[str, str]] = set()
    for from_node, to_node, key in collector.graph.edges:
        if key.edge_type == EdgeType.default:
            edges_by_kind.add((from_node.kind, to_node.kind))
    # create the ddl metadata from the kinds
    model.create_schema(list(edges_by_kind))
    # ingest the data
    parquet_conf = config.get("destinations", {}).get("parquet")
    assert parquet_conf
    parquet_path = Path(parquet_conf["path"])
    parquet_batch_size = int(parquet_conf["batch_size"])
    writer = ParquetWriter(model, parquet_path, parquet_batch_size)
    node: BaseResource
    for node in sorted(collector.graph.nodes, key=lambda n: n.kind):
        exported = prepare_node(node, collector)
        writer.insert_node(exported)
        ne_current += 1
        if ne_current % progress_update == 0:
            feedback.progress_done("sync_db", ne_current, node_edge_count, context=[collector.cloud])
    for from_node, to_node, key in collector.graph.edges:
        if key.edge_type == EdgeType.default:
            writer.insert_node({"from": from_node.chksum, "to": to_node.chksum, "type": "edge"})
            ne_current += 1
            if ne_current % progress_update == 0:
                feedback.progress_done("sync_db", ne_current, node_edge_count, context=[collector.cloud])

    writer.close()

    feedback.progress_done(collector.cloud, 1, 1)
    return collector.cloud, len(collector.graph.nodes), len(collector.graph.edges)


def collect_sql(
    collector: BaseCollectorPlugin, engine: Engine, feedback: CoreFeedback, args: Namespace
) -> Tuple[str, int, int]:
    # collect cloud data
    feedback.progress_done(collector.cloud, 0, 1)
    collector.collect()

    # group all nodes by kind
    nodes_by_kind: Dict[str, List[Json]] = defaultdict(list)
    node: BaseResource
    for node in collector.graph.nodes:
        # create an exported node with the same scheme as resotocore
        exported = prepare_node(node, collector)
        nodes_by_kind[node.kind].append(exported)

    # group all edges by kind of from/to
    edges_by_kind: Dict[Tuple[str, str], List[Json]] = defaultdict(list)
    for from_node, to_node, key in collector.graph.edges:
        if key.edge_type == EdgeType.default:
            edge_node = {"from": from_node.chksum, "to": to_node.chksum, "type": "edge"}
            edges_by_kind[(from_node.kind, to_node.kind)].append(edge_node)

    # read the kinds created from this collector
    kinds = [from_json(m, Kind) for m in collector.graph.export_model(walk_subclasses=False)]
    updater = sql_updater(Model({k.fqn: k for k in kinds}), engine)
    node_edge_count = len(collector.graph.nodes) + len(collector.graph.edges)
    ne_count = 0
    schema = f"create temp tables {engine.dialect.name}"
    syncdb = f"synchronize {engine.dialect.name}"
    feedback.progress_done(schema, 0, 1, context=[collector.cloud])
    feedback.progress_done(syncdb, 0, node_edge_count, context=[collector.cloud])
    with engine.connect() as conn:
        with conn.begin():
            # create the ddl metadata from the kinds
            updater.create_schema(conn, args, list(edges_by_kind.keys()))
            feedback.progress_done(schema, 1, 1, context=[collector.cloud])

            # insert batches of nodes by kind
            for kind, nodes in nodes_by_kind.items():
                log.info(f"Inserting {len(nodes)} nodes of kind {kind}")
                for insert in updater.insert_nodes(kind, nodes):
                    conn.execute(insert)
                ne_count += len(nodes)
                feedback.progress_done(syncdb, ne_count, node_edge_count, context=[collector.cloud])

            # insert batches of edges by from/to kind
            for from_to, nodes in edges_by_kind.items():
                log.info(f"Inserting {len(nodes)} edges from {from_to[0]} to {from_to[1]}")
                for insert in updater.insert_edges(from_to, nodes):
                    conn.execute(insert)
                ne_count += len(nodes)
                feedback.progress_done(syncdb, ne_count, node_edge_count, context=[collector.cloud])
    feedback.progress_done(collector.cloud, 1, 1)
    return collector.cloud, len(collector.graph.nodes), len(collector.graph.edges)


def show_messages(core_messages: Queue[Json], end: Event) -> None:
    info = CollectInfo()
    while not end.is_set():
        with Live(info.render(), transient=True):
            with suppress(Exception):
                info.handle_message(core_messages.get(timeout=1))
    for message in info.rendered_messages():
        rich_print(message)


def collect_from_plugins(engine: Optional[Engine], args: Namespace, sender: AnalyticsEventSender) -> None:
    # the multiprocessing manager is used to share data between processes
    mp_manager = multiprocessing.Manager()
    core_messages: Queue[Json] = mp_manager.Queue()
    feedback = CoreFeedback("cloud2sql", "collect", "collect", core_messages)
    raw_config = configure(args.config)
    sources = raw_config["sources"]
    all_collectors = collectors(sources, feedback)
    engine_name = engine.dialect.name if engine else "parquet"
    analytics = {"total": len(all_collectors), "engine": engine_name} | {name: 1 for name in all_collectors}
    end = Event()
    with ThreadPoolExecutor(max_workers=4) as executor:
        try:
            if args.show == "progress":
                executor.submit(show_messages, core_messages, end)
            futures: List[Future[Any]] = []
            for collector in all_collectors.values():
                futures.append(executor.submit(collect, collector, engine, feedback, args, raw_config))
            for future in concurrent.futures.as_completed(futures):
                name, nodes, edges = future.result()
                analytics[f"{name}_nodes"] = nodes
                analytics[f"{name}_edges"] = edges
            sender.capture("collect", **analytics)
            # when all collectors are done, we can swap all temp tables
            if engine:
                swap_tables = "Make latest snapshot available"
                feedback.progress_done(swap_tables, 0, 1)
                SqlUpdater.swap_temp_tables(engine)
                feedback.progress_done(swap_tables, 1, 1)
        except Exception as e:
            # set end and wait for live to finish, otherwise the cursor is not reset
            end.set()
            sender.capture("error", error=str(e))
            sleep(1)
            sender.flush()
            log.error("An error occurred", exc_info=e)
            print(f"Encountered Error. Giving up. {e}")
            emergency_shutdown()
        finally:
            end.set()



import logging
import uuid
from abc import abstractmethod, ABC
from datetime import datetime
from typing import Union

from requests import get
from posthog import Client

log = logging.getLogger("resoto.cloud2sql")


class AnalyticsEventSender(ABC):
    """
    Abstract event sender used to send analytics events.
    """

    @abstractmethod
    def capture(self, kind: str, **context: Union[str, int, float]) -> None:
        pass

    def flush(self) -> None:
        pass


class NoEventSender(AnalyticsEventSender):
    """
    Use this sender to not emit any events other than writing it to the log file.
    """

    def capture(self, kind: str, **context: Union[str, int, float]) -> None:
        log.debug(f"Capture event {kind}: {context}")


class PosthogEventSender(AnalyticsEventSender):
    """
    This sender is used to collect analytics events in posthog.
    """

    def __init__(self) -> None:
        self.client = Client(host="https://analytics.some.engineering", api_key="")
        self.uid = uuid.uuid4()

    def capture(self, kind: str, **context: Union[str, int, float]) -> None:
        if self.client.api_key == "":
            api_key = get("https://cdn.some.engineering/posthog/public_api_key").text.strip()
            self.client.api_key = api_key
            for consumer in self.client.consumers:
                consumer.api_key = api_key
        self.client.capture(
            distinct_id=self.uid,
            event="cloud2sql." + kind,
            properties={
                **context,
                "source": "cloud2sql",
                "run_id": self.uid,
            },
            timestamp=datetime.now(),
        )

    def flush(self) -> None:
        self.client.flush()



import logging
from abc import ABC, abstractmethod
from typing import List, Any, Type, Tuple, Dict, Iterator

from resotoclient.models import Kind, Model
from resotolib.args import Namespace
from resotolib.types import Json
from sqlalchemy import Boolean, Column, Float, Integer, JSON, MetaData, String, Table, DDL
from sqlalchemy.engine import Engine, Connection
from sqlalchemy.sql.ddl import DropTable, DropConstraint
from sqlalchemy.sql.dml import ValuesBase

from cloud2sql.util import value_in_path
from cloud2sql.schema_utils import (
    base_kinds,
    temp_prefix,
    carz_access,
    get_table_name,
    get_link_table_name,
    kind_properties,
)

log = logging.getLogger("resoto.cloud2sql")


def sql_kind_to_column_type(kind_name: str, model: Model) -> Any:  # Type[TypeEngine[Any]]
    kind = model.kinds.get(kind_name)
    if "[]" in kind_name:
        return JSON
    elif kind_name.startswith("dict"):
        return JSON
    elif kind_name == "any":
        return JSON
    elif kind_name in ("int32", "int64"):
        return Integer
    elif kind_name in "float":
        return Float
    elif kind_name in "double":
        return Float  # use Double with sqlalchemy 2
    elif kind_name in ("string", "date", "datetime", "duration"):
        return String
    elif kind_name == "boolean":
        return Boolean
    elif kind and kind.properties:  # complex kind
        return JSON
    elif kind and kind.runtime_kind is not None:  # refined simple type like enum
        return sql_kind_to_column_type(kind.runtime_kind, model)
    else:
        raise ValueError(f"Not able to handle kind {kind_name}")


class SqlUpdater(ABC):
    @abstractmethod
    def insert_nodes(self, kind: str, nodes: List[Json]) -> Iterator[ValuesBase]:
        pass

    @abstractmethod
    def insert_edges(self, from_to: Tuple[str, str], nodes: List[Json]) -> Iterator[ValuesBase]:
        pass

    @abstractmethod
    def create_schema(self, connection: Connection, args: Namespace, edges: List[Tuple[str, str]]) -> MetaData:
        pass

    @staticmethod
    def swap_temp_tables(engine: Engine) -> None:
        with engine.connect() as connection:
            with connection.begin():
                metadata = MetaData()
                metadata.reflect(connection, resolve_fks=False)

                def drop_table(tl: Table) -> None:
                    for cs in tl.foreign_key_constraints:
                        connection.execute(DropConstraint(cs))
                    connection.execute(DropTable(tl))

                for table in metadata.tables.values():
                    if table.name.startswith(temp_prefix):
                        prod_table = table.name[len(temp_prefix) :]  # noqa: E203
                        if prod_table in metadata.tables:
                            drop_table(metadata.tables[prod_table])
                        connection.execute(DDL(f"ALTER TABLE {table.name} RENAME TO {prod_table}"))
                # todo: create foreign key constraints on the final tables


class SqlDefaultUpdater(SqlUpdater):
    def __init__(self, model: Model, **kwargs: Any) -> None:
        self.model = model
        self.metadata = MetaData()
        self.table_kinds = [
            kind
            for kind in model.kinds.values()
            if kind.aggregate_root and kind.runtime_kind is None and kind.fqn not in base_kinds
        ]
        self.kind_by_id: Dict[str, str] = {}
        self.column_types_fn = kwargs.get("kind_to_column_type", sql_kind_to_column_type)
        self.insert_batch_size = kwargs.get("insert_batch_size", 5000)

    def create_schema(self, connection: Connection, args: Namespace, edges: List[Tuple[str, str]]) -> MetaData:
        log.info(f"Create schema for {len(self.table_kinds)} kinds and their relationships")

        def table_schema(kind: Kind) -> None:
            table_name = get_table_name(kind.fqn)
            if table_name not in self.metadata.tables:
                properties, _ = kind_properties(kind, self.model)
                Table(
                    get_table_name(kind.fqn),
                    self.metadata,
                    *[
                        Column("_id", String, primary_key=True),
                        *[Column(p.name, self.column_types_fn(p.kind, self.model)) for p in properties],
                    ],
                )

        def link_table_schema(from_kind: str, to_kind: str) -> None:
            from_table = get_table_name(from_kind)
            to_table = get_table_name(to_kind)
            link_table = get_link_table_name(from_kind, to_kind)
            if (
                link_table not in self.metadata.tables
                and from_table in self.metadata.tables
                and to_table in self.metadata.tables
            ):
                # defining a foreign key constraint on a table that does not exist yet, will fail
                # defining it on the current tmp table setup is possible, but this will not reset during table rename
                Table(link_table, self.metadata, Column("from_id", String), Column("to_id", String))

        def link_table_schema_from_successors(kind: Kind) -> None:
            _, successors = kind_properties(kind, self.model)
            # create link table for all linked entities
            for successor in successors:
                link_table_schema(kind.fqn, successor)

        # step 1: create tables for all kinds
        for kind in self.table_kinds:
            table_schema(kind)
        # step 2: create link tables for all kinds
        for kind in self.table_kinds:
            link_table_schema_from_successors(kind)
        # step 3: create link tables for all seen edges
        for from_kind, to_kind in edges:
            link_table_schema(from_kind, to_kind)

        # drop tables if requested
        self.metadata.drop_all(connection)

        # create the tables
        self.metadata.create_all(connection)

        return self.metadata

    def node_to_json(self, node: Json) -> Json:
        if node.get("type") == "node" and "id" in node and "reported" in node:
            reported: Json = node.get("reported", {})
            reported["_id"] = node["id"]
            reported["cloud"] = value_in_path(node, carz_access["cloud"])
            reported["account"] = value_in_path(node, carz_access["account"])
            reported["region"] = value_in_path(node, carz_access["region"])
            reported["zone"] = value_in_path(node, carz_access["zone"])
            reported.pop("kind", None)
            return reported
        elif node.get("type") == "edge" and "from" in node and "to" in node:
            return {"from_id": node["from"], "to_id": node["to"]}
        raise ValueError(f"Unknown node: {node}")

    def insert_nodes(self, kind: str, nodes: List[Json]) -> Iterator[ValuesBase]:
        # create a defaults dict with all properties set to None
        kp, _ = kind_properties(self.model.kinds[kind], self.model)
        defaults = {p.name: None for p in kp}

        if (table := self.metadata.tables.get(get_table_name(kind))) is not None:
            for batch in (nodes[i : i + self.insert_batch_size] for i in range(0, len(nodes), self.insert_batch_size)):
                converted = [defaults | self.node_to_json(node) for node in batch]
                yield table.insert().values(converted)

    def insert_edges(self, from_to: Tuple[str, str], nodes: List[Json]) -> Iterator[ValuesBase]:
        from_kind, to_kind = from_to
        table = self.metadata.tables.get(get_link_table_name(from_kind, to_kind))
        maybe_insert = table.insert() if table is not None else None
        if maybe_insert is not None:
            for batch in (nodes[i : i + self.insert_batch_size] for i in range(0, len(nodes), self.insert_batch_size)):
                converted = [self.node_to_json(node) for node in batch]
                yield maybe_insert.values(converted)


# register your updater by dialect name here
DialectUpdater: Dict[str, Type[SqlUpdater]] = {}


def sql_updater(model: Model, engine: Engine) -> SqlUpdater:
    updater_class: Type[SqlUpdater] = DialectUpdater.get(engine.dialect.name, SqlDefaultUpdater)
    updater: SqlUpdater = updater_class(model)  # type: ignore
    log.info(f"Dialect {engine.dialect.name}: Use updater class {updater_class.__name__}")
    return updater



from resotoclient.models import Kind, Model, JsObject
from typing import Dict, List, Any, NamedTuple, Optional, Tuple
import pyarrow as pa
from cloud2sql.schema_utils import (
    base_kinds,
    get_table_name,
    get_link_table_name,
    kind_properties,
    insert_node,
)
import pyarrow.parquet as pq
from pathlib import Path
from dataclasses import dataclass


class ParquetModel:
    def __init__(self, model: Model):
        self.model = model
        self.table_kinds = [
            kind
            for kind in model.kinds.values()
            if kind.aggregate_root and kind.runtime_kind is None and kind.fqn not in base_kinds
        ]
        self.schemas: Dict[str, pa.Schema] = {}

    def _parquet_type(self, kind: str) -> pa.lib.DataType:
        if kind.startswith("dict") or "[]" in kind:
            return pa.string()  # dicts and lists are converted to json strings
        elif kind == "int32":
            return pa.int32()
        elif kind == "int64":
            return pa.int64()
        elif kind == "float":
            pa.float32()
        elif kind == "double":
            return pa.float64()
        elif kind in {"string", "datetime", "date", "duration"}:
            return pa.string()
        elif kind == "boolean":
            return pa.bool_()
        else:
            return pa.string()

    def create_schema(self, edges: List[Tuple[str, str]]) -> None:
        def table_schema(kind: Kind) -> None:
            table_name = get_table_name(kind.fqn, with_tmp_prefix=False)
            if table_name not in self.schemas:
                properties, _ = kind_properties(kind, self.model)
                schema = pa.schema(
                    [
                        pa.field("_id", pa.string()),
                        *[pa.field(p.name, self._parquet_type(p.kind)) for p in properties],
                    ]
                )
                self.schemas[table_name] = schema

        def link_table_schema(from_kind: str, to_kind: str) -> None:
            from_table = get_table_name(from_kind, with_tmp_prefix=False)
            to_table = get_table_name(to_kind, with_tmp_prefix=False)
            link_table = get_link_table_name(from_kind, to_kind, with_tmp_prefix=False)
            if link_table not in self.schemas and from_table in self.schemas and to_table in self.schemas:
                schema = pa.schema(
                    [
                        pa.field("from_id", pa.string()),
                        pa.field("to_id", pa.string()),
                    ]
                )
                self.schemas[link_table] = schema

        def link_table_schema_from_successors(kind: Kind) -> None:
            _, successors = kind_properties(kind, self.model)
            # create link table for all linked entities
            for successor in successors:
                link_table_schema(kind.fqn, successor)

        # step 1: create tables for all kinds
        for kind in self.table_kinds:
            table_schema(kind)
        # step 2: create link tables for all kinds
        for kind in self.table_kinds:
            link_table_schema_from_successors(kind)
        # step 3: create link tables for all seen edges
        for from_kind, to_kind in edges:
            link_table_schema(from_kind, to_kind)

        return None


class WriteResult(NamedTuple):
    table_name: str


@dataclass
class ParquetBatch:
    rows: List[Dict[str, Any]]
    schema: pa.Schema
    writer: pq.ParquetWriter


class ParquetWriter:
    def __init__(
        self,
        model: ParquetModel,
        result_directory: Path,
        rows_per_batch: int,
    ):
        self.model = model
        self.kind_by_id: Dict[str, str] = {}
        self.batches: Dict[str, ParquetBatch] = {}
        self.rows_per_batch = rows_per_batch
        self.result_directory = result_directory

    def insert_value(self, table_name: str, values: Any) -> Optional[WriteResult]:
        if self.model.schemas.get(table_name):

            def ensure_path(path: Path) -> Path:
                path.mkdir(parents=True, exist_ok=True)
                return path

            batch = self.batches.get(
                table_name,
                ParquetBatch(
                    [],
                    self.model.schemas[table_name],
                    pq.ParquetWriter(
                        Path(ensure_path(self.result_directory), f"{table_name}.parquet"),
                        self.model.schemas[table_name],
                    ),
                ),
            )

            batch.rows.append(values)
            self.batches[table_name] = batch
            return WriteResult(table_name)
        return None

    def write_batch_bundle(self, batch: ParquetBatch) -> None:
        rows = batch.rows
        batch.rows = []
        pa_table = pa.Table.from_pylist(rows, batch.schema)
        batch.writer.write_table(pa_table)

    def insert_node(self, node: JsObject) -> None:
        result = insert_node(
            node,
            self.kind_by_id,
            self.insert_value,
            with_tmp_prefix=False,
            flatten=True,
        )
        should_write_batch = result and len(self.batches[result.table_name].rows) > self.rows_per_batch
        if result and should_write_batch:
            batch = self.batches[result.table_name]
            self.write_batch_bundle(batch)

    def close(self) -> None:
        for batch in self.batches.values():
            self.write_batch_bundle(batch)
            batch.writer.close()



from logging import getLogger
from typing import Optional
from resotolib.args import Namespace, ArgumentParser
from resotolib.logger import setup_logger
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine

from cloud2sql.analytics import PosthogEventSender, NoEventSender, AnalyticsEventSender
from cloud2sql.collect_plugins import collect_from_plugins, configure
from cloud2sql.util import db_string_from_config, check_parquet_driver

# Will fail in case snowflake is not installed - which is fine.
try:
    from cloud2sql.snowflake import SnowflakeUpdater  # noqa:F401
except ImportError:
    pass

log = getLogger("cloud2sql")


def parse_args() -> Namespace:
    parser = ArgumentParser(
        description="Collect data from cloud providers and store it in a database",
        epilog="Synchronizes cloud data to a database",
        env_args_prefix="CLOUD2SQL_",
    )
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")
    parser.add_argument("--config", help="Path to config file", required=True)
    parser.add_argument(
        "--show",
        choices=["progress", "log", "none"],
        default="progress",
        help="Output to show during the process. Default: progress",
    )
    parser.add_argument(
        "--analytics-opt-out",
        default=False,
        action="store_true",
        help="Do not send anonimized analytics data",
    )
    args = parser.parse_args()
    args.log_level = "CRITICAL" if args.show != "log" else "DEBUG" if args.debug else "INFO"
    return args  # type: ignore


def collect(engine: Optional[Engine], args: Namespace, sender: AnalyticsEventSender) -> None:
    try:
        collect_from_plugins(engine, args, sender)
    except Exception as e:
        log.error("Error during collection", e)
        print(f"Error syncing data to database: {e}")


def main() -> None:
    args = parse_args()
    sender: Optional[AnalyticsEventSender] = None
    try:
        setup_logger("resoto.cloud2sql", level=args.log_level, force=True)
        sender = NoEventSender() if args.analytics_opt_out else PosthogEventSender()
        config = configure(args.config)
        engine = None
        if next(iter(config["destinations"].keys()), None) == "parquet":
            check_parquet_driver()
        else:
            engine = create_engine(db_string_from_config(config))
        collect(engine, args, sender)
    except Exception as e:
        if args.debug:  # raise exception and show complete tracelog
            raise e
        else:
            print(f"Error syncing data to database: {e}")
    finally:
        if sender:
            sender.flush()


if __name__ == "__main__":
    main()



from resotolib.baseresources import BaseResource
from resotolib import baseresources
import inspect
from resotoclient.models import Property, Kind, JsObject, Model
from resotolib.types import Json
from typing import List, Dict, Tuple, Optional, Any, Callable, TypeVar
from cloud2sql.util import value_in_path
import json

# This set will hold the names of all "base" resources
# Since that are abstract classes, there will be no instances of them - hence we do not need a table for them.
base_kinds = {
    clazz.kind for _, clazz in inspect.getmembers(baseresources, inspect.isclass) if issubclass(clazz, BaseResource)
}

temp_prefix = "tmp_"

carz = [
    Property("cloud", "string"),
    Property("account", "string"),
    Property("region", "string"),
    Property("zone", "string"),
]
carz_access = {name: ["ancestors", name, "reported", "id"] for name in ["cloud", "account", "region", "zone"]}


def get_table_name(kind: str, with_tmp_prefix: bool = True) -> str:
    replaced = kind.replace(".", "_")
    return temp_prefix + replaced if with_tmp_prefix else replaced


def get_link_table_name(from_kind: str, to_kind: str, with_tmp_prefix: bool = True) -> str:
    # postgres table names are not allowed to be longer than 63 characters
    replaced = f"link_{get_table_name(from_kind, False)[0:25]}_{get_table_name(to_kind, False)[0:25]}"
    return temp_prefix + replaced if with_tmp_prefix else replaced


def kind_properties(kind: Kind, model: Model) -> Tuple[List[Property], List[str]]:
    visited = set()

    def base_props_not_visited(kd: Kind) -> Tuple[List[Property], List[str]]:
        if kd.fqn in visited:
            return [], []
        visited.add(kd.fqn)
        # take all properties that are not synthetic
        # also ignore the kind property, since it is available in the table name
        properties: Dict[str, Property] = {
            prop.name: prop for prop in (kd.properties or []) if prop.synthetic is None and prop.name != "kind"
        }
        defaults = kd.successor_kinds.get("default") if kd.successor_kinds else None
        successors: List[str] = defaults.copy() if defaults else []
        for kind_name in kd.bases or []:
            if ck := model.kinds.get(kind_name):  # and kind_name != "resource":
                props, succs = base_props_not_visited(ck)
                for prop in props:
                    properties[prop.name] = prop
                successors.extend(succs)
        return list(properties.values()), successors

    prs, scs = base_props_not_visited(kind)
    return prs + carz, scs


T = TypeVar("T")


def insert_node(
    node: JsObject,
    kind_by_id: Dict[str, str],
    insert_value: Callable[[str, Any], Optional[T]],
    with_tmp_prefix: bool = True,
    flatten: bool = False,
) -> Optional[T]:
    if node.get("type") == "node" and "id" in node and "reported" in node:
        reported: Json = node.get("reported", {})
        if flatten:
            for name, value in reported.items():
                if isinstance(value, (dict, list, set)):
                    reported[name] = json.dumps(value)  # convert to json in case of a nested object

        reported["_id"] = node["id"]
        reported["cloud"] = value_in_path(node, carz_access["cloud"])
        reported["account"] = value_in_path(node, carz_access["account"])
        reported["region"] = value_in_path(node, carz_access["region"])
        reported["zone"] = value_in_path(node, carz_access["zone"])
        kind = reported.pop("kind")
        kind_by_id[node["id"]] = kind
        return insert_value(get_table_name(kind, with_tmp_prefix), reported)
    elif node.get("type") == "edge" and "from" in node and "to" in node:
        from_id = node["from"]
        to_id = node["to"]
        if (from_kind := kind_by_id.get(from_id)) and (to_kind := kind_by_id.get(to_id)):
            link_table = get_link_table_name(from_kind, to_kind, with_tmp_prefix)
            return insert_value(link_table, {"from_id": from_id, "to_id": to_id})
    raise ValueError(f"Unknown node: {node}")



"""
Cloud2Sql
~~~~~~~~~~~~~~~~~~~~~~~~~~
Read infrastructure data from your cloud and export it to a SQL database.
:copyright: © 2022 Some Engineering Inc.
:license: Apache 2.0, see LICENSE for more details.
"""

__title__ = "cloud2sql"
__description__ = "Read infrastructure data from your cloud and export it to a SQL database."
__author__ = "Some Engineering Inc."
__license__ = "Apache 2.0"
__copyright__ = "Copyright © 2022 Some Engineering Inc."
__version__ = "0.6.0"



from typing import Union, List, Optional, Any

from resotolib.types import JsonElement, Json
from sqlalchemy.engine import create_engine


def value_in_path(element: JsonElement, path_or_name: Union[List[str], str]) -> Optional[Any]:
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path)

    def at_idx(current: JsonElement, idx: int) -> Optional[Any]:
        if at == idx:
            return current
        elif current is None or not isinstance(current, dict) or path[idx] not in current:
            return None
        else:
            return at_idx(current[path[idx]], idx + 1)

    return at_idx(element, 0)


def db_string_from_config(config: Json) -> str:
    destinations = config.get("destinations", {})

    if len(destinations) != 1:
        raise ValueError("Exactly one destination must be configured")

    db_type = list(destinations.keys())[0]
    db_config = destinations[db_type]
    db_type = update_db_type(db_type)
    user = db_config.get("user")
    password = db_config.get("password")
    host = db_config.get("host")
    port = db_config.get("port")
    database = db_config.get("database")
    args = db_config.get("args", {})

    db_uri = f"{db_type}://"

    if user:
        db_uri += user
        if password:
            db_uri += f":{password}"
        db_uri += "@"

    if host:
        db_uri += host
        if port:
            db_uri += f":{port}"

    if database:
        db_uri += f"/{database}"

    if len(args) > 0:
        db_uri += "?" + "&".join([f"{k}={v}" for k, v in args.items()])

    check_db_driver(db_uri)

    return db_uri


def update_db_type(db_type: str) -> str:
    if db_type == "mysql":
        db_type = "mysql+pymysql"
    elif db_type == "mariadb":
        db_type = "mariadb+pymysql"
    return db_type


def check_db_driver(db_uri: str) -> None:
    try:
        create_engine(db_uri)
    except ModuleNotFoundError:
        err = "The database type you configured is not installed. "
        if db_uri.startswith("mysql") or db_uri.startswith("mariadb"):
            err += "Please run 'pip install cloud2sql[mysql]' and try again."
        elif db_uri.startswith("postgresql"):
            err += "Please run 'pip install cloud2sql[postgresql]' and try again."
        elif db_uri.startswith("snowflake"):
            err += "Please run 'pip install cloud2sql[snowflake]' and try again."
        elif db_uri.startswith("mssql"):
            err += "Please install the pymssql package and try again."
        elif db_uri.startswith("oracle"):
            err += "Please install the cx_oracle package and try again."
        else:
            err += "Please install the required dependencies and try again."
        raise ModuleNotFoundError(err)


def check_parquet_driver() -> None:
    try:
        import pyarrow  # noqa: F401
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "The parquet format you configured is not installed. "
            "Please run 'pip install cloud2sql[parquet]' and try again."
        )



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

import cloud2sql

with open("requirements.txt") as f:
    required = f.read().splitlines()

with open("requirements-mysql.txt") as f:
    required_mysql = f.read().splitlines()

with open("requirements-postgresql.txt") as f:
    required_postgresql = f.read().splitlines()

with open("requirements-parquet.txt") as f:
    required_parquet = f.read().splitlines()

with open("requirements-snowflake.txt") as f:
    required_snowflake = f.read().splitlines()

with open("requirements-test.txt") as f:
    test_required = f.read().splitlines()

with open("README.md") as f:
    readme = f.read()

setup_requirements = [
    "pytest-runner",
]


setup(
    name=cloud2sql.__title__,
    version=cloud2sql.__version__,
    description=cloud2sql.__description__,
    python_requires=">=3.9",
    classifiers=["Programming Language :: Python :: 3"],
    entry_points={"console_scripts": ["cloud2sql=cloud2sql.__main__:main"]},
    install_requires=required,
    extras_require={
        "all": required_mysql + required_postgresql + required_snowflake + required_parquet,
        "mysql": required_mysql,
        "mariadb": required_mysql,
        "postgresql": required_postgresql,
        "snowflake": required_snowflake,
        "parquet": required_parquet,
    },
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    packages=find_packages(include=["cloud2sql", "cloud2sql.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_required,
    url="https://github.com/someengineering/cloud2sql",
)



import json
import logging
from typing import Any, List, Iterator

from resotoclient import Model
from resotoclient.models import Property
from resotolib.types import Json
from snowflake.sqlalchemy import ARRAY, OBJECT
from sqlalchemy import Integer, Float, String, Boolean, column
from sqlalchemy import select
from sqlalchemy.sql import Values
from sqlalchemy.sql.dml import ValuesBase

from cloud2sql.sql import SqlDefaultUpdater, DialectUpdater

from cloud2sql.schema_utils import kind_properties, get_table_name

log = logging.getLogger("resoto.cloud2sql.snowflake")


def kind_to_snowflake_type(kind_name: str, model: Model) -> Any:  # Type[TypeEngine[Any]]
    """
    Map internal kinds to snowflake types.
    More or less the default mapping, but with some special cases for OBJECT and ARRAY types.
    """
    kind = model.kinds.get(kind_name)
    if "[]" in kind_name:
        return ARRAY
    elif kind_name.startswith("dict"):
        return OBJECT
    elif kind_name == "any":
        return OBJECT
    elif kind_name in ("int32", "int64"):
        return Integer
    elif kind_name in "float":
        return Float
    elif kind_name in "double":
        return Float  # use Double with sqlalchemy 2
    elif kind_name in ("string", "date", "datetime", "duration"):
        return String
    elif kind_name == "boolean":
        return Boolean
    elif kind.runtime_kind is not None:  # refined simple type like enum
        return kind_to_snowflake_type(kind.runtime_kind, model)
    elif kind.properties:  # complex kind
        return OBJECT
    else:
        raise ValueError(f"Not able to handle kind {kind_name}")


class SnowflakeUpdater(SqlDefaultUpdater):
    """
    This updater synchronizes resource data to snowflake https://www.snowflake.com
    Snowflake needs special handling, since it does not support default json or array types.
    It also does not understand json or array types as bind parameters.
    This updater handles those shortcomings by using special insert statements.
    """

    def __init__(self, model: Model, **args: Any) -> None:
        super().__init__(model, **args)
        self.column_types_fn = kind_to_snowflake_type

    def insert_nodes(self, kind: str, nodes: List[Json]) -> Iterator[ValuesBase]:
        kp, _ = kind_properties(self.model.kinds[kind], self.model)
        kind_props = [Property("_id", "string")] + kp
        select_array = []
        column_definitions = []
        prop_is_json = {}

        # Inserting structured data into Snowflake requires a bit of work. General scheme:
        # insert into TBL(col_string, col_json) SELECT column1, parse_json(column2) from values('a', '{"b":1}');
        # All json and array elements need to be json encoded and parsed on the server side again.
        for num, prop in enumerate(kind_props):
            name = f"column{num+1}"
            select_array.append(prop.name)
            snowflake_kind = kind_to_snowflake_type(prop.kind, self.model)
            if snowflake_kind in (ARRAY, OBJECT):
                column_definitions.append(column(f"parse_json({name})", is_literal=True))
                prop_is_json[prop.name] = True
            else:
                column_definitions.append(column(name))

        def values_tuple(node: Json) -> List[Any]:
            nj = self.node_to_json(node)
            # make sure to use the same order as in select_array
            return [json.dumps(nj.get(p.name)) if prop_is_json.get(p.name) else nj.get(p.name) for p in kind_props]

        if (table := self.metadata.tables.get(get_table_name(kind))) is not None:
            for batch in (nodes[i : i + self.insert_batch_size] for i in range(0, len(nodes), self.insert_batch_size)):
                converted = [values_tuple(node) for node in batch]
                yield table.insert().from_select(select_array, select(Values(*column_definitions).data(converted)))


# register this updater for the snowflake dialect, when snowflake is installed
DialectUpdater["snowflake"] = SnowflakeUpdater



import shutil
from typing import Optional, Dict, Any, List

from resotolib.core.progress import ProgressTree, Progress
from resotolib.types import Json
from rich.console import Group
from rich.markdown import Markdown
from rich.text import Text
from rich.tree import Tree as RichTree


class CollectInfo:
    def __init__(self) -> None:
        self.progress = ProgressTree("Progress")
        self.messages: List[Json] = []
        self.info_count: int = 0
        self.error_count: int = 0

    def handle_message(self, message: Json) -> Json:
        if message["kind"] == "action_progress":
            self.progress.add_progress(Progress.from_json(message["data"]["progress"]))
        else:
            if message["data"]["level"] == "info":
                self.info_count += 1
            else:
                self.error_count += 1
            self.messages.append(message)
        return message

    def render(self, max_height: Optional[int] = None) -> Any:
        max_height = max_height or shutil.get_terminal_size(fallback=(80, 25))[1]
        res: Any = progress_to_table(self.progress, max_height)
        if self.messages:
            res = Group(Markdown(f"- {self.error_count} errors\n- {self.info_count} warnings"), res)
        return res

    def rendered_messages(self) -> List[Any]:
        def markup(m: Json) -> Text:
            level = "[red]ERROR[/red]" if m["data"]["level"] == "error" else "[yellow]WARNING[/yellow]"
            return Text.from_markup(f"{level}: {m['data']['message']}")

        return [markup(m) for m in self.messages]


def progress_to_table(progress: ProgressTree, max_height: int) -> RichTree:
    remaining = max_height

    def render_progress(pg: Progress) -> Any:
        pi = pg.overall_progress()
        emoji = ":white_check_mark:" if pi.done else ":arrows_counterclockwise:"
        text = ("done" if pi.done else "in_progress") if pi.total == 1 else f"{pi.percentage}%"
        return Text.from_markup(f"{emoji} {pg.name} [dim]{text}[/dim]")

    def walk_node(nid: str, node: Dict[str, Any], rt: Optional[RichTree] = None) -> RichTree:
        nonlocal remaining
        if remaining < 1:
            return rt or RichTree(progress.name)
        dn = node[nid]
        data = dn["data"]
        node_content = render_progress(data) if data is not None else Text(nid)
        sub = RichTree(progress.name) if rt is None else rt.add(node_content)
        remaining -= 1

        for child in dn.get("children", []):
            for nid, _ in child.items():
                walk_node(nid, child, sub)
        return sub

    return walk_node(progress.sub_tree.root, progress.sub_tree.to_dict(with_data=True))
